# -*- coding: utf-8 -*-
"""The evaluator base class in agentscope."""
import asyncio
from typing import Callable, Awaitable, Coroutine, Any

from .._benchmark_base import BenchmarkBase
from .._evaluator._evaluator_base import EvaluatorBase
from .._solution import SolutionOutput
from .._task import Task
from .._evaluator_storage import EvaluatorStorageBase


def _check_ray_available() -> None:
    """Check if ray is available and raise ImportError if not."""
    try:
        import ray  # noqa  # pylint: disable=unused-import
    except ImportError as e:
        raise ImportError(
            "Ray is not installed. Please install it with `pip install ray` "
            "to use the RayEvaluator.",
        ) from e


# Create a conditional decorator for ray.remote
def _ray_remote_decorator(cls: Any) -> Any:
    """
    Conditional ray.remote decorator that only applies when ray is available.
    """
    try:
        import ray

        return ray.remote(cls)
    except ImportError:
        return cls


@_ray_remote_decorator
class RayEvaluationActor:
    """
    Actor class for running evaluation with ray remote.
    """

    @staticmethod
    async def run(
        storage: EvaluatorStorageBase,
        task: Task,
        repeat_id: str,
        solution_output: SolutionOutput,
    ) -> None:
        """
        Run the evaluation for a task and solution result.

        Args:
            storage (EvaluatorStorageBase): Evaluator storage.
            task (Task): Task to be evaluated.
            repeat_id (str): Repeat ID
            solution_output (SolutionOutput): output data after execute agents.
        """
        evaluation_results = await task.evaluate(solution_output)
        # store the evaluation result
        for result in evaluation_results:
            storage.save_evaluation_result(
                task_id=task.id,
                repeat_id=repeat_id,
                evaluation=result,
            )


@_ray_remote_decorator
class RaySolutionActor:
    """
    Actor class for running agent solutions with ray remote.
    """

    def __init__(self, n_workers: int = 1):
        self.eval_actor = RayEvaluationActor.options(
            max_concurrency=n_workers,
        ).remote()

    async def run(
        self,
        storage: EvaluatorStorageBase,
        repeat_id: str,
        task: Task,
        solution: Callable[
            [Task, Callable],
            Coroutine[Any, Any, SolutionOutput],
        ],
    ) -> None:
        """Generate a solution to a task and evaluate.

        Args:
            storage (EvaluatorStorageBase): Evaluator storage.
            repeat_id (str): Repeat ID.
            task (Task): Task to be evaluated.
            solution
                (Callable[[Task, Callable], Awaitable[SolutionOutput, Any]]):
                callable function to execute agents and generate results.
        """
        if storage.solution_result_exists(task.id, repeat_id):
            # Obtain from storage
            solution_result = storage.get_solution_result(
                task.id,
                repeat_id,
            )

        else:
            # Run the solution
            solution_result = await solution(
                task,
                storage.get_agent_pre_print_hook(
                    task.id,
                    repeat_id,
                ),
            )

            storage.save_solution_result(
                task.id,
                repeat_id,
                solution_result,
            )

        # Evaluate the solution with the
        futures = []
        for metric in task.metrics:
            if not storage.evaluation_result_exists(
                task.id,
                repeat_id,
                metric.name,
            ):
                futures.append(
                    self.eval_actor.run.remote(
                        storage,
                        task,
                        repeat_id,
                        solution_result,
                    ),
                )
        if futures:
            await asyncio.gather(*futures)


class RayEvaluator(EvaluatorBase):
    """The ray-based evaluator that supports distributed and parallel
    evaluation."""

    def __init__(
        self,
        name: str,
        benchmark: BenchmarkBase,
        n_repeat: int,
        storage: EvaluatorStorageBase,
        n_workers: int,
    ) -> None:
        """Initialize the evaluator."""
        super().__init__(
            name=name,
            benchmark=benchmark,
            n_repeat=n_repeat,
            storage=storage,
        )

        # Check ray availability early
        _check_ray_available()

        assert isinstance(benchmark, BenchmarkBase)

        assert n_repeat >= 1, "n_repeat must be at least 1"

        assert n_workers >= 1, "n_workers must be at least 1"

        self.benchmark = benchmark
        self.n_repeat = n_repeat
        self.n_workers = n_workers

    async def run(
        self,
        solution: Callable[
            [Task, Callable],
            Awaitable[SolutionOutput] | SolutionOutput,
        ],
    ) -> None:
        """Run the ray-based distributed and parallel evaluation, and get the
        results.

        Args:
            solution (`Callable[[Task], SolutionOutput]`):
                A sync or async function that takes a `Task` instance as input
                and returns a `SolutionOutput` instance.
        """

        await self._save_evaluation_meta()

        futures = []
        solution_actor = RaySolutionActor.options(
            max_concurrency=self.n_workers,
        ).remote(n_workers=self.n_workers)
        for repeat_id in range(self.n_repeat):
            for task in self.benchmark:
                futures.append(
                    solution_actor.run.remote(
                        self.storage,
                        str(repeat_id),
                        task,
                        solution,
                    ),
                )
        if futures:
            await asyncio.gather(*futures)

        await self.aggregate()



# -*- coding: utf-8 -*-
"""Task memory implementation using ReMe library.

This module provides a task memory implementation that integrates
with the ReMe library to learn from execution trajectories and
retrieve relevant task experiences.

Requirements:
    Python 3.12 or greater is required to use ReMe.
"""
from typing import Any

from ._reme_long_term_memory_base import ReMeLongTermMemoryBase
from ..._logging import logger
from ...message import Msg, TextBlock
from ...tool import ToolResponse


class ReMeTaskLongTermMemory(ReMeLongTermMemoryBase):
    """Task memory implementation using ReMe library.

    Task memory learns from execution trajectories and provides
    retrieval of relevant task experiences.

    Requirements:
        Python 3.12 or greater is required to use ReMe.
    """

    async def record_to_memory(
        self,
        thinking: str,
        content: list[str],
        **kwargs: Any,
    ) -> ToolResponse:
        """Record task execution experiences and learnings.

        Record task execution experiences and learnings to long-term
        memory.

        Use this function to save valuable task-related knowledge that
        can help with future similar tasks. This enables learning from
        experience and improving over time.

        When to record:

        - After solving technical problems or completing tasks
        - When discovering useful techniques or approaches
        - After implementing solutions with specific steps
        - When learning best practices or important lessons

        What to record: Be detailed and actionable. Include:

        - Task description and context
        - Step-by-step execution details
        - Specific techniques and methods used
        - Results, outcomes, and effectiveness
        - Lessons learned and considerations

        Args:
            thinking (`str`):
                Your reasoning about why this task experience is valuable
                and what makes it worth remembering for future reference.
            content (`list[str]`):
                List of specific task insights to remember. Each string
                should be a clear, actionable piece of information.
                Examples: ["Add indexes on WHERE clause columns to speed
                up queries", "Use EXPLAIN ANALYZE to identify missing
                indexes"].
            **kwargs (`Any`):
                Additional keyword arguments. Can include 'score' (float)
                to indicate the quality/success of this approach
                (default: 1.0).

        Returns:
            `ToolResponse`:
                Confirmation message indicating successful memory
                recording.
        """
        logger.info(
            "[ReMeTaskMemory] Entering record_to_memory - "
            "thinking: %s, content: %s, kwargs: %s",
            thinking,
            content,
            kwargs,
        )

        if not self._app_started:
            raise RuntimeError(
                "ReMeApp context not started. "
                "Please use 'async with' to initialize the app.",
            )

        try:
            # Prepare messages for task memory recording
            messages = []

            # Add thinking as a user message if provided
            if thinking:
                messages.append(
                    {
                        "role": "user",
                        "content": thinking,
                    },
                )

            # Add content items as user-assistant pairs
            for item in content:
                messages.append(
                    {
                        "role": "user",
                        "content": item,
                    },
                )
                # Add a simple assistant acknowledgment
                messages.append(
                    {
                        "role": "assistant",
                        "content": "Task information recorded.",
                    },
                )

            result = await self.app.async_execute(
                name="summary_task_memory",
                workspace_id=self.workspace_id,
                trajectories=[
                    {
                        "messages": messages,
                        "score": kwargs.pop("score", 1.0),
                    },
                ],
                **kwargs,
            )

            # Extract metadata if available
            summary_text = (
                f"Successfully recorded {len(content)} task memory/memories."
            )

            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=summary_text,
                    ),
                ],
                metadata={"result": result},
            )

        except Exception as e:
            logger.exception("Error recording task memory: %s", str(e))
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Error recording task memory: {str(e)}",
                    ),
                ],
            )

    async def retrieve_from_memory(
        self,
        keywords: list[str],
        **kwargs: Any,
    ) -> ToolResponse:
        """Search and retrieve relevant task experiences.

        Search and retrieve relevant task experiences from long-term
        memory.

        IMPORTANT: You should call this function BEFORE attempting to
        solve problems or answer technical questions. This ensures you
        leverage experiences and proven solutions rather than
        starting from scratch.

        Use this when:
        - Asked to solve a technical problem or implement a solution
        - Asked for recommendations, best practices, or approaches
        - Asked "what do you know about...?" or "have you seen this
          before?"
        - Dealing with tasks that may be similar to experiences
        - Need to recall specific techniques or methods

        Benefits of retrieving first:
        - Learn from past successes and mistakes
        - Provide more accurate, battle-tested solutions
        - Avoid reinventing the wheel
        - Give consistent, informed recommendations

        Args:
            keywords (`list[str]`):
                Keywords describing the task or problem domain. Be
                specific and use technical terms. Examples:
                ["database optimization", "slow queries"], ["API design",
                "rate limiting"], ["code refactoring", "Python"].
            **kwargs (`Any`):
                Additional keyword arguments. Can include 'top_k' (int)
                to specify number of experiences to retrieve
                (default: 3).

        Returns:
            `ToolResponse`:
                Retrieved task experiences and learnings. If no relevant
                experiences found, you'll receive a message indicating
                that.
        """
        logger.info(
            "[ReMeTaskMemory] Entering retrieve_from_memory - "
            "keywords: %s, kwargs: %s",
            keywords,
            kwargs,
        )

        if not self._app_started:
            raise RuntimeError(
                "ReMeApp context not started. "
                "Please use 'async with' to initialize the app.",
            )

        try:
            results = []

            # Search for each keyword
            top_k = kwargs.get("top_k", 3)
            for keyword in keywords:
                result = await self.app.async_execute(
                    name="retrieve_task_memory",
                    workspace_id=self.workspace_id,
                    query=keyword,
                    top_k=top_k,
                    **kwargs,
                )

                # Extract the answer from the result
                answer = result.get("answer", "")
                if answer:
                    results.append(f"Keyword '{keyword}':\n{answer}")

            # Combine all results
            if results:
                combined_text = "\n\n".join(results)
            else:
                combined_text = (
                    "No task experiences found for the given keywords."
                )

            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=combined_text,
                    ),
                ],
            )

        except Exception as e:
            logger.exception("Error retrieving task memory: %s", str(e))
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Error retrieving task memory: {str(e)}",
                    ),
                ],
            )

    async def record(
        self,
        msgs: list[Msg | None],
        **kwargs: Any,
    ) -> None:
        """Record the content to the task memory.

        This method converts AgentScope messages to ReMe's format and
        records them as a task execution trajectory.

        Args:
            msgs (`list[Msg | None]`):
                The messages to record to memory.
            **kwargs (`Any`):
                Additional keyword arguments for the recording.
                Can include 'score' (float) for trajectory scoring
                (default: 1.0).
        """
        if isinstance(msgs, Msg):
            msgs = [msgs]

        # Filter out None
        msg_list = [_ for _ in msgs if _]
        if not msg_list:
            return

        if not all(isinstance(_, Msg) for _ in msg_list):
            raise TypeError(
                "The input messages must be a list of Msg objects.",
            )

        if not self._app_started:
            raise RuntimeError(
                "ReMeApp context not started. "
                "Please use 'async with' to initialize the app.",
            )

        try:
            # Convert AgentScope messages to ReMe format
            messages = []
            for msg in msg_list:
                # Extract content as string
                if isinstance(msg.content, str):
                    content_str = msg.content
                elif isinstance(msg.content, list):
                    # Join content blocks into a single string
                    content_parts = []
                    for block in msg.content:
                        if isinstance(block, dict) and "text" in block:
                            content_parts.append(block["text"])
                        elif isinstance(block, dict) and "thinking" in block:
                            content_parts.append(block["thinking"])
                    content_str = "\n".join(content_parts)
                else:
                    content_str = str(msg.content)

                messages.append(
                    {
                        "role": msg.role,
                        "content": content_str,
                    },
                )

            # Extract score from kwargs if provided, default to 1.0
            score = kwargs.pop("score", 1.0)

            await self.app.async_execute(
                name="summary_task_memory",
                workspace_id=self.workspace_id,
                trajectories=[
                    {
                        "messages": messages,
                        "score": score,
                    },
                ],
                **kwargs,
            )

        except Exception as e:
            # Log the error but don't raise to maintain compatibility
            logger.exception(
                "Error recording messages to task memory: %s",
                str(e),
            )
            import warnings

            warnings.warn(
                f"Error recording messages to task memory: {str(e)}",
            )

    async def retrieve(
        self,
        msg: Msg | list[Msg] | None,
        **kwargs: Any,
    ) -> str:
        """Retrieve relevant task experiences from memory.

        Args:
            msg (`Msg | list[Msg] | None`):
                The message to search for relevant task experiences.
            **kwargs (`Any`):
                Additional keyword arguments.

        Returns:
            `str`:
                The retrieved task experiences as a string.
        """
        if msg is None:
            return ""

        if isinstance(msg, Msg):
            msg = [msg]

        if not isinstance(msg, list) or not all(
            isinstance(_, Msg) for _ in msg
        ):
            raise TypeError(
                "The input message must be a Msg or a list of Msg objects.",
            )

        if not self._app_started:
            raise RuntimeError(
                "ReMeApp context not started. "
                "Please use 'async with' to initialize the app.",
            )

        try:
            # Only use the last message's content for retrieval
            last_msg = msg[-1]
            query = ""

            if isinstance(last_msg.content, str):
                query = last_msg.content
            elif isinstance(last_msg.content, list):
                # Extract text from content blocks
                content_parts = []
                for block in last_msg.content:
                    if isinstance(block, dict) and "text" in block:
                        content_parts.append(block["text"])
                    elif isinstance(block, dict) and "thinking" in block:
                        content_parts.append(block["thinking"])
                query = "\n".join(content_parts)

            if not query:
                return ""

            # Retrieve using the query from the last message
            top_k = kwargs.get("top_k", 3)
            result = await self.app.async_execute(
                name="retrieve_task_memory",
                workspace_id=self.workspace_id,
                query=query,
                top_k=top_k,
                **kwargs,
            )

            return result.get("answer", "")

        except Exception as e:
            logger.exception("Error retrieving task memory: %s", str(e))
            import warnings

            warnings.warn(f"Error retrieving task memory: {str(e)}")
            return ""



# -*- coding: utf-8 -*-
"""The formatter module in agentscope."""

from ._formatter_base import FormatterBase
from ._truncated_formatter_base import TruncatedFormatterBase
from ._dashscope_formatter import (
    DashScopeChatFormatter,
    DashScopeMultiAgentFormatter,
)
from ._anthropic_formatter import (
    AnthropicChatFormatter,
    AnthropicMultiAgentFormatter,
)
from ._openai_formatter import (
    OpenAIChatFormatter,
    OpenAIMultiAgentFormatter,
)
from ._gemini_formatter import (
    GeminiChatFormatter,
    GeminiMultiAgentFormatter,
)
from ._ollama_formatter import (
    OllamaChatFormatter,
    OllamaMultiAgentFormatter,
)
from ._deepseek_formatter import (
    DeepSeekChatFormatter,
    DeepSeekMultiAgentFormatter,
)

__all__ = [
    "FormatterBase",
    "TruncatedFormatterBase",
    "DashScopeChatFormatter",
    "DashScopeMultiAgentFormatter",
    "OpenAIChatFormatter",
    "OpenAIMultiAgentFormatter",
    "AnthropicChatFormatter",
    "AnthropicMultiAgentFormatter",
    "GeminiChatFormatter",
    "GeminiMultiAgentFormatter",
    "OllamaChatFormatter",
    "OllamaMultiAgentFormatter",
    "DeepSeekChatFormatter",
    "DeepSeekMultiAgentFormatter",
]



# -*- coding: utf-8 -*-
"""The in-memory plan storage class."""
from collections import OrderedDict

from ._plan_model import Plan
from ._storage_base import PlanStorageBase


class InMemoryPlanStorage(PlanStorageBase):
    """In-memory plan storage."""

    def __init__(self) -> None:
        """Initialize the in-memory plan storage."""
        super().__init__()
        self.plans = OrderedDict()

        # Support historical plan serialization and deserialization
        self.register_state(
            "plans",
            lambda plans: {k: v.model_dump() for k, v in plans.items()},
            lambda json_data: OrderedDict(
                (k, Plan.model_validate(v)) for k, v in json_data.items()
            ),
        )

    async def add_plan(self, plan: Plan, override: bool = True) -> None:
        """Add a plan to the storage.

        Args:
            plan (`Plan`):
                The plan to be added.
            override (`bool`, defaults to `True`):
                Whether to override the existing plan with the same ID.
        """
        if plan.id in self.plans and not override:
            raise ValueError(
                f"Plan with id {plan.id} already exists.",
            )
        self.plans[plan.id] = plan

    async def delete_plan(self, plan_id: str) -> None:
        """Delete a plan from the storage.

        Args:
            plan_id (`str`):
                The ID of the plan to be deleted.
        """
        self.plans.pop(plan_id, None)

    async def get_plans(self) -> list[Plan]:
        """Get all plans from the storage.

        Returns:
            `list[Plan]`:
                A list of all plans in the storage.
        """
        return list(self.plans.values())

    async def get_plan(self, plan_id: str) -> Plan | None:
        """Get a plan by its ID.

        Args:
            plan_id (`str`):
                The ID of the plan to be retrieved.

        Returns:
            `Plan | None`:
                The plan with the specified ID, or None if not found.
        """
        return self.plans.get(plan_id, None)



# -*- coding: utf-8 -*-
"""The dialogue memory class"""

from typing import Union, Iterable, Any

from ._memory_base import MemoryBase
from ..message import Msg


class InMemoryMemory(MemoryBase):
    """The in-memory memory class for storing messages."""

    def __init__(
        self,
    ) -> None:
        """Initialize the in-memory memory object."""
        super().__init__()
        self.content: list[Msg] = []

    def state_dict(self) -> dict:
        """Convert the current memory into JSON data format."""
        return {
            "content": [_.to_dict() for _ in self.content],
        }

    def load_state_dict(
        self,
        state_dict: dict,
        strict: bool = True,
    ) -> None:
        """Load the memory from JSON data.

        Args:
            state_dict (`dict`):
                The state dictionary to load, which should have a "content"
                field.
            strict (`bool`, defaults to `True`):
                If `True`, raises an error if any key in the module is not
                found in the state_dict. If `False`, skips missing keys.
        """
        self.content = []
        for data in state_dict["content"]:
            data.pop("type", None)
            self.content.append(Msg.from_dict(data))

    async def size(self) -> int:
        """The size of the memory."""
        return len(self.content)

    async def retrieve(self, *args: Any, **kwargs: Any) -> None:
        """Retrieve items from the memory."""
        raise NotImplementedError(
            "The retrieve method is not implemented in "
            f"{self.__class__.__name__} class.",
        )

    async def delete(self, index: Union[Iterable, int]) -> None:
        """Delete the specified item by index(es).

        Args:
            index (`Union[Iterable, int]`):
                The index to delete.
        """
        if isinstance(index, int):
            index = [index]

        invalid_index = [_ for _ in index if 0 > _ or _ >= len(self.content)]

        if invalid_index:
            raise IndexError(
                f"The index {invalid_index} does not exist.",
            )

        self.content = [
            _ for idx, _ in enumerate(self.content) if idx not in index
        ]

    async def add(
        self,
        memories: Union[list[Msg], Msg, None],
        allow_duplicates: bool = False,
    ) -> None:
        """Add message into the memory.

        Args:
            memories (`Union[list[Msg], Msg, None]`):
                The message to add.
            allow_duplicates (`bool`, defaults to `False`):
                If allow adding duplicate messages (with the same id) into
                the memory.
        """
        if memories is None:
            return

        if isinstance(memories, Msg):
            memories = [memories]

        if not isinstance(memories, list):
            raise TypeError(
                f"The memories should be a list of Msg or a single Msg, "
                f"but got {type(memories)}.",
            )

        for msg in memories:
            if not isinstance(msg, Msg):
                raise TypeError(
                    f"The memories should be a list of Msg or a single Msg, "
                    f"but got {type(msg)}.",
                )

        if not allow_duplicates:
            existing_ids = [_.id for _ in self.content]
            memories = [_ for _ in memories if _.id not in existing_ids]
        self.content.extend(memories)

    async def get_memory(self) -> list[Msg]:
        """Get the memory content."""
        return self.content

    async def clear(self) -> None:
        """Clear the memory content."""
        self.content = []



# -*- coding: utf-8 -*-
"""The tool module in agentscope."""

from ._response import ToolResponse
from ._coding import (
    execute_python_code,
    execute_shell_command,
)
from ._text_file import (
    view_text_file,
    write_text_file,
    insert_text_file,
)
from ._multi_modality import (
    dashscope_text_to_image,
    dashscope_text_to_audio,
    dashscope_image_to_text,
    openai_text_to_image,
    openai_text_to_audio,
    openai_edit_image,
    openai_create_image_variation,
    openai_image_to_text,
    openai_audio_to_text,
)
from ._toolkit import Toolkit

__all__ = [
    "Toolkit",
    "ToolResponse",
    "execute_python_code",
    "execute_shell_command",
    "view_text_file",
    "write_text_file",
    "insert_text_file",
    "dashscope_text_to_image",
    "dashscope_text_to_audio",
    "dashscope_image_to_text",
    "openai_text_to_image",
    "openai_text_to_audio",
    "openai_edit_image",
    "openai_create_image_variation",
    "openai_image_to_text",
    "openai_audio_to_text",
]



# -*- coding: utf-8 -*-
"""The MCP module in AgentScope, that provides fine-grained control over
the MCP servers."""

from ._client_base import MCPClientBase
from ._mcp_function import MCPToolFunction
from ._stateful_client_base import StatefulClientBase
from ._stdio_stateful_client import StdIOStatefulClient
from ._http_stateless_client import HttpStatelessClient
from ._http_stateful_client import HttpStatefulClient


__all__ = [
    "MCPToolFunction",
    "MCPClientBase",
    "StatefulClientBase",
    "StdIOStatefulClient",
    "HttpStatelessClient",
    "HttpStatefulClient",
]



# -*- coding: utf-8 -*-
"""The reme memory module."""

from ._reme_personal_long_term_memory import ReMePersonalLongTermMemory
from ._reme_task_long_term_memory import ReMeTaskLongTermMemory
from ._reme_tool_long_term_memory import ReMeToolLongTermMemory

__all__ = [
    "ReMePersonalLongTermMemory",
    "ReMeTaskLongTermMemory",
    "ReMeToolLongTermMemory",
]



# -*- coding: utf-8 -*-
"""The state module in agentscope."""

import json
from collections import OrderedDict
from dataclasses import dataclass
from typing import Callable, Any, Optional

from ..types import JSONSerializableObject


@dataclass
class _JSONSerializeFunction:
    to_json: Optional[Callable[[Any], Any]] = None
    """The function converting the original data to JSON data."""
    load_json: Optional[Callable[[Any], Any]] = None
    """The function converting the JSON data to original data."""


class StateModule:
    """The state module class in agentscope to support nested state
    serialization and deserialization."""

    def __init__(self) -> None:
        """Initialize the state module."""
        self._module_dict = OrderedDict()
        self._attribute_dict = OrderedDict()

    def __setattr__(self, key: str, value: Any) -> None:
        """Set attributes and record state modules."""
        if isinstance(value, StateModule):
            if not hasattr(self, "_module_dict"):
                raise AttributeError(
                    f"Call the super().__init__() method within the "
                    f"constructor of {self.__class__.__name__} before setting "
                    f"any attributes.",
                )
            self._module_dict[key] = value
        super().__setattr__(key, value)

    def __delattr__(self, key: str) -> None:
        """Delete attributes and remove from state modules."""
        if key in self._module_dict:
            self._module_dict.pop(key)
        if key in self._attribute_dict:
            self._attribute_dict.pop(key)
        super().__delattr__(key)

    def state_dict(self) -> dict:
        """Get the state dictionary of the module, including the nested
        state modules and registered attributes.

        Returns:
            `dict`:
                A dictionary that keys are attribute names and values are
                the state of the attribute.
        """
        state = {}
        for key in self._module_dict:
            attr = getattr(self, key, None)
            if isinstance(attr, StateModule):
                state[key] = attr.state_dict()

        for key in self._attribute_dict:
            attr = getattr(self, key)
            to_json_function = self._attribute_dict[key].to_json
            if to_json_function is not None:
                state[key] = to_json_function(attr)
            else:
                state[key] = attr

        return state

    def load_state_dict(self, state_dict: dict, strict: bool = True) -> None:
        """Load the state dictionary into the module.

        Args:
            state_dict (`dict`):
                The state dictionary to load.
            strict (`bool`, defaults to `True`):
                If `True`, raises an error if any key in the module is not
                found in the state_dict. If `False`, skips missing keys.
        """
        for key in self._module_dict:
            if key not in state_dict:
                if strict:
                    raise KeyError(
                        f"Key '{key}' not found in state_dict. Ensure that "
                        f"the state_dict contains all required keys.",
                    )
                continue
            self._module_dict[key].load_state_dict(state_dict[key])

        for key in self._attribute_dict:
            if key not in state_dict:
                if strict:
                    raise KeyError(
                        f"Key '{key}' not found in state_dict. Ensure that "
                        f"the state_dict contains all required keys.",
                    )
                continue
            from_json_func = self._attribute_dict[key].load_json
            if from_json_func is not None:
                setattr(self, key, from_json_func(state_dict[key]))
            else:
                setattr(self, key, state_dict[key])

    def register_state(
        self,
        attr_name: str,
        custom_to_json: Callable[[Any], JSONSerializableObject] | None = None,
        custom_from_json: Callable[[JSONSerializableObject], Any]
        | None = None,
    ) -> None:
        """Register an attribute to be tracked as a state variable.

        Args:
            attr_name (`str`):
                The name of the attribute to register.
            custom_to_json (`Callable[[Any], JSONSerializableObject] | None`, \
            optional):
                A custom function to convert the attribute to a
                JSON-serializable format. If not provided, `json.dumps` will
                be used.
            custom_from_json (`Callable[[JSONSerializableObject], Any] | None`\
            , defaults to `None`):
                A custom function to convert the JSON dictionary back to the
                original attribute format.
        """
        attr = getattr(self, attr_name)

        if custom_to_json is None:
            # Make sure the attribute is JSON serializable natively
            try:
                json.dumps(attr)
            except Exception as e:
                raise TypeError(
                    f"Attribute '{attr_name}' is not JSON serializable. "
                    "Please provide a custom function to convert the "
                    "attribute to a JSON-serializable format.",
                ) from e

        if attr_name in self._module_dict:
            raise ValueError(
                f"Attribute `{attr_name}` is already registered as a module. ",
            )

        self._attribute_dict[attr_name] = _JSONSerializeFunction(
            to_json=custom_to_json,
            load_json=custom_from_json,
        )



# -*- coding: utf-8 -*-
"""The PDF reader to read and chunk PDF files."""
import hashlib
from typing import Literal

from ._reader_base import ReaderBase
from ._text_reader import TextReader
from .._document import Document


class PDFReader(ReaderBase):
    """The PDF reader that splits text into chunks by a fixed chunk size."""

    def __init__(
        self,
        chunk_size: int = 512,
        split_by: Literal["char", "sentence", "paragraph"] = "sentence",
    ) -> None:
        """Initialize the text reader.

        Args:
            chunk_size (`int`, default to 512):
                The size of each chunk, in number of characters.
            split_by (`Literal["char", "sentence", "paragraph"]`, default to \
            "sentence"):
                The unit to split the text, can be "char", "sentence", or
                "paragraph". The "sentence" option is implemented using the
                "nltk" library, which only supports English text.
        """
        if chunk_size <= 0:
            raise ValueError(
                f"The chunk_size must be positive, got {chunk_size}",
            )

        if split_by not in ["char", "sentence", "paragraph"]:
            raise ValueError(
                "The split_by must be one of 'char', 'sentence' or "
                f"'paragraph', got {split_by}",
            )

        self.chunk_size = chunk_size
        self.split_by = split_by

        # To avoid code duplication, we use TextReader to do the chunking.
        self._text_reader = TextReader(
            self.chunk_size,
            self.split_by,
        )

    async def __call__(
        self,
        pdf_path: str,
    ) -> list[Document]:
        """Read a PDF file, split it into chunks, and return a list of
        Document objects.

        Args:
            pdf_path (`str`):
                The input PDF file path.
        """
        try:
            from pypdf import PdfReader
        except ImportError as e:
            raise ImportError(
                "Please install pypdf to use the PDF reader. "
                "You can install it by `pip install pypdf`.",
            ) from e

        reader = PdfReader(pdf_path)

        gather_texts = []
        for page in reader.pages:
            gather_texts.append(page.extract_text())

        doc_id = hashlib.sha256(pdf_path.encode("utf-8")).hexdigest()

        docs = await self._text_reader("\n\n".join(gather_texts))
        for doc in docs:
            doc.id = doc_id

        return docs

    def get_doc_id(self, pdf_path: str) -> str:
        """Get the document ID. This function can be used to check if the
        doc_id already exists in the knowledge base."""
        return hashlib.sha256(pdf_path.encode("utf-8")).hexdigest()



# -*- coding: utf-8 -*-
"""The base MCP stateful client class in AgentScope, that provides basic
 functionality for stateful MCP clients."""
from abc import ABC
from contextlib import AsyncExitStack
from typing import List

import mcp
from mcp import ClientSession

from ._client_base import MCPClientBase
from ._mcp_function import MCPToolFunction
from .._logging import logger


class StatefulClientBase(MCPClientBase, ABC):
    """The base class for stateful MCP clients in AgentScope, which maintains
    the session state across multiple tool calls.

    The developers should use `connect()` and `close()` methods to manage
    the client lifecycle.
    """

    is_connected: bool
    """If connected to the MCP server"""

    def __init__(self, name: str) -> None:
        """Initialize the stateful MCP client.

        Args:
            name (`str`):
                The name to identify the MCP server, which should be unique
                across the MCP servers.
        """

        super().__init__(name=name)

        self.client = None
        self.stack = None
        self.session = None
        self.is_connected = False

        # Cache the tools to avoid fetching them multiple times
        self._cached_tools = None

    async def connect(self) -> None:
        """Connect to MCP server."""
        if self.is_connected:
            raise RuntimeError(
                "The MCP server is already connected. Call close() "
                "before connecting again.",
            )

        self.stack = AsyncExitStack()

        try:
            context = await self.stack.enter_async_context(
                self.client,
            )
            read_stream, write_stream = context[0], context[1]
            self.session = ClientSession(read_stream, write_stream)
            await self.stack.enter_async_context(self.session)
            await self.session.initialize()

            self.is_connected = True
            logger.info("MCP client connected.")
        except Exception:
            await self.stack.aclose()
            self.stack = None
            raise

    async def close(self) -> None:
        """Clean up the MCP client resources. You must call this method when
        your application is done."""
        if not self.is_connected:
            raise RuntimeError(
                "The MCP server is not connected. Call connect() before "
                "closing.",
            )

        try:
            await self.stack.aclose()
            logger.info("MCP client closed.")
        except Exception as e:
            logger.warning("Error during MCP client cleanup: %s", e)
        finally:
            self.stack = None
            self.session = None
            self.is_connected = False

    async def list_tools(self) -> List[mcp.types.Tool]:
        """Get all available tools from the server.

        Returns:
            `mcp.types.ListToolsResult`:
                A list of available MCP tools.
        """
        self._validate_connection()

        res = await self.session.list_tools()

        # Cache the tools for later use
        self._cached_tools = res.tools
        return res.tools

    async def get_callable_function(
        self,
        func_name: str,
        wrap_tool_result: bool = True,
    ) -> MCPToolFunction:
        """Get an async tool function from the MCP server by its name, so
        that you can call it directly, wrap it into your own function, or
        anyway you like.

        .. note:: Currently, only the text, image, and audio results are
         supported in this function.

        Args:
            func_name (`str`):
                The name of the tool function to get.
            wrap_tool_result (`bool`):
                Whether to wrap the tool result into agentscope's
                `ToolResponse` object. If `False`, the raw result type
                `mcp.types.CallToolResult` will be returned.

        Returns:
            `MCPToolFunction`:
                A callable async function that returns either
                `mcp.types.CallToolResult` or `ToolResponse` when called.
        """
        self._validate_connection()

        if self._cached_tools is None:
            await self.list_tools()

        target_tool = None
        for tool in self._cached_tools:
            if tool.name == func_name:
                target_tool = tool
                break

        if target_tool is None:
            raise ValueError(
                f"Tool '{func_name}' not found in the MCP server",
            )

        return MCPToolFunction(
            mcp_name=self.name,
            tool=target_tool,
            wrap_tool_result=wrap_tool_result,
            session=self.session,
        )

    def _validate_connection(self) -> None:
        """Validate the connection to the MCP server."""
        if not self.is_connected:
            raise RuntimeError(
                "The connection is not established. Call connect() "
                "before using the client.",
            )

        if not self.session:
            raise RuntimeError(
                "The session is not initialized. Call connect() "
                "before using the client.",
            )



# -*- coding: utf-8 -*-
"""The ACE benchmark metric implementations in AgentScope."""

from .._solution import SolutionOutput
from .._metric_base import MetricBase, MetricResult, MetricType


class ACEProcessAccuracy(MetricBase):
    """The ace benchmark process accuracy metric."""

    def __init__(
        self,
        mile_stone: list[str],
    ) -> None:
        """Initialize the AceBench process accuracy metric."""
        super().__init__(
            name="process_accuracy",
            metric_type=MetricType.NUMERICAL,
            description="The AceBench Agent eval process accuracy metric.",
        )
        self.mile_stone = mile_stone

    async def __call__(
        self,
        solution: SolutionOutput,
    ) -> MetricResult:
        """Calculate the metric result."""

        # Turn the tool use block sequence into ACEBench format
        # e.g. func(arg1='dfd', arg2=44)
        gathered_trajectory = []
        for tool_call in solution.trajectory:
            if tool_call.get("type") == "tool_use":
                function_name = tool_call.get("name")
                kwargs = tool_call.get("input")

                gathered_kwargs = []
                for key, value in kwargs.items():
                    if isinstance(value, str):
                        gathered_kwargs.append(
                            f"{key}='{value}'",
                        )

                    else:
                        gathered_kwargs.append(
                            f"{key}={value}",
                        )

                kwargs_str = ", ".join(gathered_kwargs)
                gathered_trajectory.append(
                    f"[{function_name}({kwargs_str})]",
                )

        for stone in self.mile_stone:
            if stone not in gathered_trajectory:
                return MetricResult(
                    name=self.name,
                    result=0,
                    message=f"Error: Missing milestone '{stone}' in "
                    "the given trajectory.",
                )

        return MetricResult(
            name=self.name,
            result=1,
            message="Success",
        )


class ACEAccuracy(MetricBase):
    """The ace benchmark metric"""

    def __init__(
        self,
        state: list[dict],
    ) -> None:
        """Initialize the _metric object."""
        super().__init__(
            "accuracy",
            MetricType.NUMERICAL,
            "The AceBench Agent eval accuracy metric.",
        )
        self.state = state

    async def __call__(
        self,
        solution: SolutionOutput,
    ) -> MetricResult:
        """Calculate the metric result."""
        # Check if the solution matches the ground truth
        if not isinstance(solution.output, list):
            raise ValueError("Ground truth state must be a list.")

        # Handle the typos in ACEBench dataset
        gathered_state = {}
        for item in self.state:
            for key, value in item.items():
                if key.endswith("API"):
                    key = key.replace("API", "Api")
                elif key.endswith("rpi"):
                    key = key.replace("pi", "Api")
                gathered_state[key] = value

        gathered_output = {}
        for item in solution.output:
            for key, value in item.items():
                gathered_output[key] = value

        if not set(gathered_state.keys()).issubset(gathered_output.keys()):
            raise ValueError(
                "Missing keys in solution output compared to state, "
                f"ground truth keys: {gathered_state.keys()}, "
                f"solution keys: {gathered_output.keys()}",
            )

        for key, value in gathered_state.items():
            if value != gathered_output.get(key):
                return MetricResult(
                    name=self.name,
                    result=0,
                    message=(
                        f"Error: Mismatch in key '{key}':"
                        f"\n{value}\n{gathered_output.get(key)}"
                    ),
                )

        return MetricResult(
            name=self.name,
            result=1,
            message="Success: All keys match",
        )



# -*- coding: utf-8 -*-
"""The plan module in AgentScope."""
from ._plan_model import (
    SubTask,
    Plan,
)
from ._plan_notebook import (
    DefaultPlanToHint,
    PlanNotebook,
)
from ._storage_base import PlanStorageBase
from ._in_memory_storage import InMemoryPlanStorage

__all__ = [
    "SubTask",
    "Plan",
    "DefaultPlanToHint",
    "PlanNotebook",
    "PlanStorageBase",
    "InMemoryPlanStorage",
]



# -*- coding: utf-8 -*-
"""The MCP stateful HTTP client module in AgentScope."""
from typing import Any, Literal

from mcp.client.sse import sse_client
from mcp.client.streamable_http import streamablehttp_client

from ._stateful_client_base import StatefulClientBase


class HttpStatefulClient(StatefulClientBase):
    """The stateful sse/streamable HTTP MCP client implementation in
    AgentScope.

    .. tip:: The stateful client is recommended for MCP servers that need to
     maintain session states, e.g. web browsers or other interactive
     MCP servers.

    .. note:: The stateful client will maintain one session across multiple
     tool calls, until the client is closed by explicitly calling the
     `close()` method.

    .. note:: When multiple HttpStatefulClient instances are connected,
     they should be closed following the Last In First Out (LIFO) principle
     to avoid potential errors. Always close the most recently registered
     client first, then work backwards to the first one.
     For more details, please refer to this `issue
     <https://github.com/modelcontextprotocol/python-sdk/issues/577>`_.
    """

    def __init__(
        self,
        name: str,
        transport: Literal["streamable_http", "sse"],
        url: str,
        headers: dict[str, str] | None = None,
        timeout: float = 30,
        sse_read_timeout: float = 60 * 5,
        **client_kwargs: Any,
    ) -> None:
        """Initialize the streamable HTTP MCP client.

        Args:
            name (`str`):
                The name to identify the MCP server, which should be unique
                across the MCP servers.
            transport (`Literal["streamable_http", "sse"]`):
                The transport type of MCP server. Generally, the URL of sse
                transport should end with `/sse`, while the streamable HTTP
                URL ends with `/mcp`.
            url (`str`):
                The URL to the MCP server.
            headers (`dict[str, str] | None`, optional):
                Additional headers to include in the HTTP request.
            timeout (`float`, optional):
                The timeout for the HTTP request in seconds. Defaults to 30.
            sse_read_timeout (`float`, optional):
                The timeout for reading Server-Sent Events (SSE) in seconds.
                Defaults to 300 (5 minutes).
            **client_kwargs (`Any`):
                The additional keyword arguments to pass to the streamable
                HTTP client.
        """
        super().__init__(name=name)

        assert transport in ["streamable_http", "sse"]
        self.transport = transport

        if self.transport == "streamable_http":
            self.client = streamablehttp_client(
                url=url,
                headers=headers,
                timeout=timeout,
                sse_read_timeout=sse_read_timeout,
                **client_kwargs,
            )
        else:
            self.client = sse_client(
                url=url,
                headers=headers,
                timeout=timeout,
                sse_read_timeout=sse_read_timeout,
                **client_kwargs,
            )



# -*- coding: utf-8 -*-
"""The tracing decorators for agent, formatter, toolkit, chat and embedding
models."""
import inspect
from functools import wraps
from typing import (
    Generator,
    AsyncGenerator,
    Callable,
    Any,
    Coroutine,
    TypeVar,
    TYPE_CHECKING,
)

import aioitertools

from ._attributes import _serialize_to_str
from .. import _config
from ..embedding._embedding_base import EmbeddingModelBase
from ..model._model_base import ChatModelBase
from .._logging import logger
from ._types import SpanKind, SpanAttributes

if TYPE_CHECKING:
    from ..agent import AgentBase
    from ..formatter import FormatterBase
    from ..tool import (
        Toolkit,
        ToolResponse,
    )
    from ..message import (
        Msg,
        ToolUseBlock,
    )
    from ..embedding import EmbeddingResponse
    from ..model import ChatResponse
    from opentelemetry.trace import Span
else:
    Toolkit = "Toolkit"
    ToolResponse = "ToolResponse"
    Msg = "Msg"
    ToolUseBlock = "ToolUseBlock"
    EmbeddingResponse = "EmbeddingResponse"
    ChatResponse = "ChatResponse"
    Span = "Span"


T = TypeVar("T")


def _check_tracing_enabled() -> bool:
    """Check if the OpenTelemetry tracer is initialized in AgentScope with an
    endpoint.

    TODO: We expect an OpenTelemetry official interface to check if the
     tracer is initialized. Leaving this function here as a temporary
     solution.
    """
    return _config.trace_enabled


def _trace_sync_generator_wrapper(
    res: Generator[T, None, None],
    span: Span,
) -> Generator[T, None, None]:
    """Trace the sync generator output with OpenTelemetry."""

    import opentelemetry

    has_error = False

    try:
        last_chunk = None
        for chunk in res:
            last_chunk = chunk
            yield chunk
    except Exception as e:
        has_error = True
        span.set_status(opentelemetry.trace.StatusCode.ERROR, str(e))
        span.record_exception(e)
        raise e from None

    finally:
        if not has_error:
            # Set the last chunk as output
            span.set_attributes(
                {
                    SpanAttributes.OUTPUT: _serialize_to_str(last_chunk),
                },
            )
            span.set_status(opentelemetry.trace.StatusCode.OK)
        span.end()


async def _trace_async_generator_wrapper(
    res: AsyncGenerator[T, None],
    span: Span,
) -> AsyncGenerator[T, None]:
    """Trace the async generator output with OpenTelemetry.

    Args:
        res (`AsyncGenerator[T, None]`):
            The generator or async generator to be traced.
        span (`Span`):
            The OpenTelemetry span to be used for tracing.

    Yields:
        `T`:
            The output of the async generator.
    """
    import opentelemetry

    has_error = False

    try:
        last_chunk = None
        async for chunk in aioitertools.iter(res):
            last_chunk = chunk
            yield chunk

    except Exception as e:
        has_error = True
        span.set_status(opentelemetry.trace.StatusCode.ERROR, str(e))
        span.record_exception(e)
        raise e from None

    finally:
        if not has_error:
            # Set the last chunk as output
            span.set_attributes(
                {
                    SpanAttributes.OUTPUT: _serialize_to_str(last_chunk),
                },
            )
            span.set_status(opentelemetry.trace.StatusCode.OK)
        span.end()


def trace(
    name: str,
) -> Callable:
    """A generic tracing decorator for synchronous and asynchronous functions.

    Args:
        name (`str`):
            The name of the span to be created.

    Returns:
        `Callable`:
            Returns a decorator that wraps the given function with
            OpenTelemetry tracing.
    """

    def decorator(
        func: Callable,
    ) -> Callable:
        """A decorator that wraps the given function with OpenTelemetry tracing

        Args:
            func (`Callable`):
                The function to be traced, which can be sync or async function,
                and returns an object or a generator.

        Returns:
            `Callable`:
                A wrapper function that traces the function call and handles
                input/output and exceptions.
        """
        # Async function
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def wrapper(
                *args: Any,
                **kwargs: Any,
            ) -> Any:
                """The wrapper function for tracing the sync function call."""
                if not _check_tracing_enabled():
                    return await func(*args, **kwargs)

                import opentelemetry

                tracer = opentelemetry.trace.get_tracer(__name__)

                attributes = {
                    SpanAttributes.SPAN_KIND: SpanKind.COMMON,
                    SpanAttributes.PROJECT_RUN_ID: _serialize_to_str(
                        _config.run_id,
                    ),
                    SpanAttributes.INPUT: _serialize_to_str(
                        {
                            "args": args,
                            "kwargs": kwargs,
                        },
                    ),
                    SpanAttributes.META: _serialize_to_str({}),
                }

                with tracer.start_as_current_span(
                    name=name,
                    attributes=attributes,
                    end_on_exit=False,
                ) as span:
                    try:
                        res = await func(*args, **kwargs)

                        # If generator or async generator
                        if isinstance(res, AsyncGenerator):
                            return _trace_async_generator_wrapper(res, span)
                        if isinstance(res, Generator):
                            return _trace_sync_generator_wrapper(res, span)

                        # non-generator result
                        span.set_attributes(
                            {SpanAttributes.OUTPUT: _serialize_to_str(res)},
                        )
                        span.set_status(opentelemetry.trace.StatusCode.OK)
                        span.end()
                        return res

                    except Exception as e:
                        span.set_status(
                            opentelemetry.trace.StatusCode.ERROR,
                            str(e),
                        )
                        span.record_exception(e)
                        span.end()
                        raise e from None

            return wrapper

        # Sync function
        @wraps(func)
        def sync_wrapper(
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            """The wrapper function for tracing the sync function call."""
            if not _check_tracing_enabled():
                return func(*args, **kwargs)

            import opentelemetry

            tracer = opentelemetry.trace.get_tracer(__name__)

            attributes = {
                SpanAttributes.SPAN_KIND: SpanKind.COMMON,
                SpanAttributes.PROJECT_RUN_ID: _serialize_to_str(
                    _config.run_id,
                ),
                SpanAttributes.INPUT: _serialize_to_str(
                    {
                        "args": args,
                        "kwargs": kwargs,
                    },
                ),
                SpanAttributes.META: _serialize_to_str({}),
            }

            with tracer.start_as_current_span(
                name=name,
                attributes=attributes,
                end_on_exit=False,
            ) as span:
                try:
                    res = func(*args, **kwargs)

                    # If generator or async generator
                    if isinstance(res, AsyncGenerator):
                        return _trace_async_generator_wrapper(res, span)
                    if isinstance(res, Generator):
                        return _trace_sync_generator_wrapper(res, span)

                    # non-generator result
                    span.set_attributes(
                        {SpanAttributes.OUTPUT: _serialize_to_str(res)},
                    )
                    span.set_status(opentelemetry.trace.StatusCode.OK)
                    span.end()
                    return res

                except Exception as e:
                    span.set_status(
                        opentelemetry.trace.StatusCode.ERROR,
                        str(e),
                    )
                    span.record_exception(e)
                    span.end()
                    raise e from None

        return sync_wrapper

    return decorator


def trace_toolkit(
    func: Callable[
        ...,
        Coroutine[Any, Any, AsyncGenerator[ToolResponse, None]],
    ],
) -> Callable[..., Coroutine[Any, Any, AsyncGenerator[ToolResponse, None]]]:
    """Trace the toolkit `call_tool_function` method with OpenTelemetry."""

    @wraps(func)
    async def wrapper(
        self: Toolkit,
        tool_call: ToolUseBlock,
    ) -> AsyncGenerator[ToolResponse, None]:
        """The wrapper function for tracing the toolkit call_tool_function
        method."""
        if not _check_tracing_enabled():
            return await func(self, tool_call=tool_call)

        import opentelemetry

        tracer = opentelemetry.trace.get_tracer(__name__)

        # Prepare the attributes for the span
        attributes = {
            SpanAttributes.SPAN_KIND: _serialize_to_str(SpanKind.TOOL),
            SpanAttributes.PROJECT_RUN_ID: _serialize_to_str(_config.run_id),
            SpanAttributes.INPUT: _serialize_to_str(
                {
                    "tool_call": tool_call,
                },
            ),
            SpanAttributes.META: _serialize_to_str(
                {
                    **tool_call,
                },
            ),
        }

        with tracer.start_as_current_span(
            f"{func.__name__}",
            attributes=attributes,
            end_on_exit=False,
        ) as span:
            try:
                # Call the toolkit function
                res = await func(self, tool_call=tool_call)

                # The result must be an AsyncGenerator of ToolResponse objects
                return _trace_async_generator_wrapper(res, span)

            except Exception as e:
                span.set_status(
                    opentelemetry.trace.StatusCode.ERROR,
                    str(e),
                )
                span.record_exception(e)
                span.end()
                raise e from None

    return wrapper


def trace_reply(
    func: Callable[..., Coroutine[Any, Any, Msg]],
) -> Callable[..., Coroutine[Any, Any, Msg]]:
    """Trace the agent reply call with OpenTelemetry.

    Args:
        func (`Callable[..., Coroutine[Any, Any, Msg]]`):
            The agent async reply function to be traced.

    Returns:
        `Callable[..., Coroutine[Any, Any, Msg]]`:
            A wrapper function that traces the agent reply call and handles
            input/output and exceptions.
    """

    @wraps(func)
    async def wrapper(
        self: "AgentBase",
        *args: Any,
        **kwargs: Any,
    ) -> Msg:
        """The wrapper function for tracing the agent reply function call."""
        if not _check_tracing_enabled():
            return await func(self, *args, **kwargs)

        from ..agent import AgentBase

        if not isinstance(self, AgentBase):
            logger.warning(
                "Skipping tracing for %s as the first argument"
                "is not an instance of AgentBase, but %s",
                func.__name__,
                type(self),
            )
            return await func(self, *args, **kwargs)

        import opentelemetry

        tracer = opentelemetry.trace.get_tracer(__name__)

        # Prepare the attributes for the span
        agent_name = self.name if hasattr(self, "name") else None
        attributes = {
            SpanAttributes.SPAN_KIND: _serialize_to_str(SpanKind.AGENT),
            SpanAttributes.PROJECT_RUN_ID: _serialize_to_str(_config.run_id),
            SpanAttributes.INPUT: _serialize_to_str(
                {
                    "args": args,
                    "kwargs": kwargs,
                },
            ),
            SpanAttributes.META: _serialize_to_str(
                {
                    "id": self.id,
                    "name": agent_name,
                },
            ),
        }

        with tracer.start_as_current_span(
            f"{self.__class__.__name__}.{func.__name__}",
            attributes=attributes,
            end_on_exit=False,
        ) as span:
            try:
                # Call the agent reply function
                res = await func(self, *args, **kwargs)

                # Set the output attribute
                span.set_attributes(
                    {SpanAttributes.OUTPUT: _serialize_to_str(res)},
                )
                span.set_status(opentelemetry.trace.StatusCode.OK)
                span.end()
                return res

            except Exception as e:
                span.set_status(
                    opentelemetry.trace.StatusCode.ERROR,
                    str(e),
                )
                span.record_exception(e)
                span.end()
                raise e from None

    return wrapper


def trace_embedding(
    func: Callable[..., Coroutine[Any, Any, EmbeddingResponse]],
) -> Callable[..., Coroutine[Any, Any, EmbeddingResponse]]:
    """Trace the embedding call with OpenTelemetry."""

    @wraps(func)
    async def wrapper(
        self: EmbeddingModelBase,
        *args: Any,
        **kwargs: Any,
    ) -> EmbeddingResponse:
        """The wrapper function for tracing the embedding call."""
        if not _check_tracing_enabled():
            return await func(self, *args, **kwargs)

        if not isinstance(self, EmbeddingModelBase):
            logger.warning(
                "Skipping tracing for %s as the first argument"
                "is not an instance of EmbeddingModelBase, but %s",
                func.__name__,
                type(self),
            )
            return await func(self, *args, **kwargs)

        import opentelemetry

        tracer = opentelemetry.trace.get_tracer(__name__)

        # Prepare the attributes for the span
        attributes = {
            SpanAttributes.SPAN_KIND: _serialize_to_str(SpanKind.EMBEDDING),
            SpanAttributes.PROJECT_RUN_ID: _serialize_to_str(_config.run_id),
            SpanAttributes.INPUT: _serialize_to_str(
                {
                    "args": args,
                    "kwargs": kwargs,
                },
            ),
            SpanAttributes.META: _serialize_to_str(
                {
                    "model_name": self.model_name,
                },
            ),
        }

        with tracer.start_as_current_span(
            f"{self.__class__.__name__}.{func.__name__}",
            attributes=attributes,
            end_on_exit=False,
        ) as span:
            try:
                # Call the embedding function
                res = await func(self, *args, **kwargs)

                # Set the output attribute
                span.set_attributes(
                    {SpanAttributes.OUTPUT: _serialize_to_str(res)},
                )
                span.set_status(opentelemetry.trace.StatusCode.OK)
                span.end()
                return res

            except Exception as e:
                span.set_status(
                    opentelemetry.trace.StatusCode.ERROR,
                    str(e),
                )
                span.record_exception(e)
                span.end()
                raise e from None

    return wrapper


def trace_format(
    func: Callable[..., Coroutine[Any, Any, list[dict]]],
) -> Callable[..., Coroutine[Any, Any, list[dict]]]:
    """Trace the format function of the formatter with OpenTelemetry.

    Args:
        func (`Callable[..., Coroutine[Any, Any, list[dict]]]`):
            The async format function to be traced.

    Returns:
        `Callable[..., Coroutine[Any, Any, list[dict]]]`:
            An async wrapper function that traces the format call and handles
            input/output and exceptions.
    """

    @wraps(func)
    async def wrapper(
        self: "FormatterBase",
        *args: Any,
        **kwargs: Any,
    ) -> list[dict]:
        """Wrap the formatter __call__ method with OpenTelemetry tracing."""
        if not _check_tracing_enabled():
            return await func(self, *args, **kwargs)

        from ..formatter import FormatterBase

        if not isinstance(self, FormatterBase):
            logger.warning(
                "Skipping tracing for %s as the first argument"
                "is not an instance of FormatterBase, but %s",
                func.__name__,
                type(self),
            )
            return await func(self, *args, **kwargs)

        import opentelemetry

        tracer = opentelemetry.trace.get_tracer(__name__)

        # Prepare the attributes for the span
        attributes = {
            SpanAttributes.SPAN_KIND: _serialize_to_str(SpanKind.FORMATTER),
            SpanAttributes.PROJECT_RUN_ID: _serialize_to_str(_config.run_id),
            SpanAttributes.INPUT: _serialize_to_str(
                {
                    "args": args,
                    "kwargs": kwargs,
                },
            ),
            SpanAttributes.META: _serialize_to_str({}),
        }

        with tracer.start_as_current_span(
            f"{self.__class__.__name__}.{func.__name__}",
            attributes=attributes,
            end_on_exit=False,
        ) as span:
            try:
                # Call the formatter function
                res = await func(self, *args, **kwargs)

                # Set the output attribute
                span.set_attributes(
                    {SpanAttributes.OUTPUT: _serialize_to_str(res)},
                )
                span.set_status(opentelemetry.trace.StatusCode.OK)
                span.end()
                return res

            except Exception as e:
                span.set_status(
                    opentelemetry.trace.StatusCode.ERROR,
                    str(e),
                )
                span.record_exception(e)
                span.end()
                raise e from None

    return wrapper


def trace_llm(
    func: Callable[
        ...,
        Coroutine[
            Any,
            Any,
            ChatResponse | AsyncGenerator[ChatResponse, None],
        ],
    ],
) -> Callable[
    ...,
    Coroutine[Any, Any, ChatResponse | AsyncGenerator[ChatResponse, None]],
]:
    """Trace the LLM call with OpenTelemetry.

    Args:
        func (`Callable`):
            The function to be traced, which should be a coroutine that
            returns either a `ChatResponse` or an `AsyncGenerator`
            of `ChatResponse`.

    Returns:
        `Callable`:
            A wrapper function that traces the LLM call and handles
            input/output and exceptions.
    """

    @wraps(func)
    async def async_wrapper(
        self: ChatModelBase,
        *args: Any,
        **kwargs: Any,
    ) -> ChatResponse | AsyncGenerator[ChatResponse, None]:
        """The wrapper function for tracing the LLM call."""
        if not _check_tracing_enabled():
            return await func(self, *args, **kwargs)

        if not isinstance(self, ChatModelBase):
            logger.warning(
                "Skipping tracing for %s as the first argument"
                "is not an instance of ChatModelBase, but %s",
                func.__name__,
                type(self),
            )
            return await func(self, *args, **kwargs)

        import opentelemetry

        tracer = opentelemetry.trace.get_tracer(__name__)

        # Prepare the attributes for the span
        attributes = {
            SpanAttributes.SPAN_KIND: _serialize_to_str(SpanKind.LLM),
            SpanAttributes.PROJECT_RUN_ID: _serialize_to_str(_config.run_id),
            SpanAttributes.INPUT: _serialize_to_str(
                {
                    "args": args,
                    "kwargs": kwargs,
                },
            ),
            SpanAttributes.META: _serialize_to_str(
                {
                    "model_name": self.model_name,
                    "stream": self.stream,
                },
            ),
        }

        # Begin the llm call span
        with tracer.start_as_current_span(
            f"{self.__class__.__name__}.__call__",
            attributes=attributes,
            end_on_exit=False,
        ) as span:
            try:
                # Must be an async calling
                res = await func(self, *args, **kwargs)

                # If the result is a AsyncGenerator
                if isinstance(res, AsyncGenerator):
                    return _trace_async_generator_wrapper(res, span)

                # non-generator result
                span.set_attributes(
                    {SpanAttributes.OUTPUT: _serialize_to_str(res)},
                )
                span.set_status(opentelemetry.trace.StatusCode.OK)
                span.end()
                return res

            except Exception as e:
                span.set_status(
                    opentelemetry.trace.StatusCode.ERROR,
                    str(e),
                )
                span.record_exception(e)
                span.end()
                raise e from None

    return async_wrapper



# -*- coding: utf-8 -*-
"""Long-term memory implementation using mem0 library.

This module provides a long-term memory implementation that integrates
with the mem0 library to provide persistent memory storage and retrieval
capabilities for AgentScope agents.
"""
import json
from typing import Any, TYPE_CHECKING
from importlib import metadata


from pydantic import field_validator

from ..embedding import EmbeddingModelBase
from ._long_term_memory_base import LongTermMemoryBase
from ..message import Msg, TextBlock
from ..model import ChatModelBase
from ..tool import ToolResponse


if TYPE_CHECKING:
    from mem0.configs.base import MemoryConfig
    from mem0.vector_stores.configs import VectorStoreConfig
else:
    MemoryConfig = Any
    VectorStoreConfig = Any


def _create_agentscope_config_classes() -> tuple:
    """Create custom config classes for agentscope providers."""
    from mem0.embeddings.configs import EmbedderConfig
    from mem0.llms.configs import LlmConfig

    class _ASLlmConfig(LlmConfig):
        """Custom LLM config class that updates the validate_config method.

        Attention: in mem0, the validate_config hardcodes the provider, so we
        need to override the validate_config method to support the agentscope
        providers. We will follow up with the mem0 to improve this.
        """

        @field_validator("config")
        @classmethod
        def validate_config(cls, v: Any, values: Any) -> Any:
            """Validate the LLM configuration."""
            from mem0.utils.factory import LlmFactory

            provider = values.data.get("provider")
            if provider in LlmFactory.provider_to_class:
                return v
            raise ValueError(f"Unsupported LLM provider: {provider}")

    class _ASEmbedderConfig(EmbedderConfig):
        """Custom embedder config class that updates the validate_config
        method."""

        @field_validator("config")
        @classmethod
        def validate_config(cls, v: Any, values: Any) -> Any:
            """Validate the embedder configuration."""
            from mem0.utils.factory import EmbedderFactory

            provider = values.data.get("provider")
            if provider in EmbedderFactory.provider_to_class:
                return v
            raise ValueError(f"Unsupported Embedder provider: {provider}")

    return _ASLlmConfig, _ASEmbedderConfig


class Mem0LongTermMemory(LongTermMemoryBase):
    """A class that implements the LongTermMemoryBase interface using mem0."""

    def __init__(
        self,
        agent_name: str | None = None,
        user_name: str | None = None,
        run_name: str | None = None,
        model: ChatModelBase | None = None,
        embedding_model: EmbeddingModelBase | None = None,
        vector_store_config: VectorStoreConfig | None = None,
        mem0_config: MemoryConfig | None = None,
        default_memory_type: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Mem0LongTermMemory instance

        Args:
            agent_name (`str | None`, optional):
                The name of the agent. Default is None.
            user_name (`str | None`, optional):
                The name of the user. Default is None.
            run_name (`str | None`, optional):
                The name of the run/session. Default is None.

        .. note::
            1. At least one of `agent_name`, `user_name`, or `run_name` is
               required.
            2. During memory recording, these parameters become metadata
               for the stored memories.
            3. **Important**: mem0 will extract memories from messages
               containing role of "user" by default. If you want to
               extract memories from messages containing role of
               "assistant", you need to provide `agent_name`.
            4. During memory retrieval, only memories with matching
               metadata values will be returned.


            model (`ChatModelBase | None`, optional):
                The chat model to use for the long-term memory. If
                mem0_config is provided, this will override the LLM
                configuration. If mem0_config is None, this is required.
            embedding_model (`EmbeddingModelBase | None`, optional):
                The embedding model to use for the long-term memory. If
                mem0_config is provided, this will override the embedder
                configuration. If mem0_config is None, this is required.
            vector_store_config (`VectorStoreConfig | None`, optional):
                The vector store config to use for the long-term memory.
                If mem0_config is provided, this will override the vector store
                configuration. If mem0_config is None and this is not
                provided, defaults to Qdrant with on_disk=True.
            mem0_config (`MemoryConfig | None`, optional):
                The mem0 config to use for the long-term memory.
                If provided, individual
                model/embedding_model/vector_store_config parameters will
                override the corresponding configurations in mem0_config. If
                None, a new MemoryConfig will be created using the provided
                parameters.
            default_memory_type (`str | None`, optional):
                The type of memory to use. Default is None, to create a
                semantic memory.

        Raises:
            `ValueError`:
                If `mem0_config` is None and either `model` or
                `embedding_model` is None.
        """
        super().__init__()

        try:
            import mem0
            from mem0.configs.llms.base import BaseLlmConfig
            from mem0.utils.factory import LlmFactory, EmbedderFactory
            from packaging import version

            # Check mem0 version
            current_version = metadata.version("mem0ai")
            is_mem0_version_low = version.parse(
                current_version,
            ) <= version.parse("0.1.115")

            # Register the agentscope providers with mem0

            EmbedderFactory.provider_to_class[
                "agentscope"
            ] = "agentscope.memory._mem0_utils.AgentScopeEmbedding"
            if is_mem0_version_low:
                # For mem0 version <= 0.1.115, use the old style
                LlmFactory.provider_to_class[
                    "agentscope"
                ] = "agentscope.memory._mem0_utils.AgentScopeLLM"
            else:
                # For mem0 version > 0.1.115, use the new style
                LlmFactory.provider_to_class["agentscope"] = (
                    "agentscope.memory._mem0_utils.AgentScopeLLM",
                    BaseLlmConfig,
                )

        except ImportError as e:
            raise ImportError(
                "Please install the mem0 library by `pip install mem0ai`",
            ) from e

        # Create the custom config classes for agentscope providers dynamically
        _ASLlmConfig, _ASEmbedderConfig = _create_agentscope_config_classes()

        if agent_name is None and user_name is None and run_name is None:
            raise ValueError(
                "at least one of agent_name, user_name, and run_name is "
                "required",
            )

        # Store agent and user identifiers for memory management
        self.agent_id = agent_name
        self.user_id = user_name
        self.run_id = run_name

        # Configuration logic: Handle mem0_config parameter
        if mem0_config is not None:
            # Case 1: mem0_config is provided - override specific
            # configurations if individual params are given

            # Override LLM configuration if model is provided
            if model is not None:
                mem0_config.llm = _ASLlmConfig(
                    provider="agentscope",
                    config={"model": model},
                )

            # Override embedder configuration if embedding_model is provided
            if embedding_model is not None:
                mem0_config.embedder = _ASEmbedderConfig(
                    provider="agentscope",
                    config={"model": embedding_model},
                )

            # Override vector store configuration if vector_store_config is
            # provided
            if vector_store_config is not None:
                mem0_config.vector_store = vector_store_config

        else:
            # Case 2: mem0_config is not provided - create new configuration
            # from individual parameters

            # Validate that required parameters are provided
            if model is None or embedding_model is None:
                raise ValueError(
                    "model and embedding_model are required if mem0_config "
                    "is not provided",
                )

            # Create new MemoryConfig with provided LLM and embedder
            mem0_config = mem0.configs.base.MemoryConfig(
                llm=_ASLlmConfig(
                    provider="agentscope",
                    config={"model": model},
                ),
                embedder=_ASEmbedderConfig(
                    provider="agentscope",
                    config={"model": embedding_model},
                ),
            )

            # Set vector store configuration
            if vector_store_config is not None:
                # Use provided vector store configuration
                mem0_config.vector_store = vector_store_config
            else:
                # Use default Qdrant configuration with on-disk storage for
                # persistence set on_disk to True to enable persistence,
                # otherwise it will be in memory only
                on_disk = kwargs.get("on_disk", True)
                mem0_config.vector_store = (
                    mem0.vector_stores.configs.VectorStoreConfig(
                        config={"on_disk": on_disk},
                    )
                )

        # Initialize the async memory instance with the configured settings
        self.long_term_working_memory = mem0.AsyncMemory(mem0_config)

        # Store the default memory type for future use
        self.default_memory_type = default_memory_type

    async def record_to_memory(
        self,
        thinking: str,
        content: list[str],
        **kwargs: Any,
    ) -> ToolResponse:
        """Use this function to record important information that you may
        need later. The target content should be specific and concise, e.g.
        who, when, where, do what, why, how, etc.

        Args:
            thinking (`str`):
                Your thinking and reasoning about what to record.
            content (`list[str]`):
                The content to remember, which is a list of strings.
        """
        # Multi-strategy recording approach to ensure content persistence:
        #
        # This method employs a three-tier fallback strategy to maximize
        # successful memory recording:
        #
        # 1. Primary: Record as "user" role message
        #    - This is the default approach for capturing user-related
        #      content
        #    - Mem0 extracts and infers memories from messages containing
        #      role of "user"
        #
        # 2. Fallback (if agent_id exists): Record as "assistant" role
        #    message
        #    - Triggered when primary recording yields no results
        #    - In this case, mem0 will use the AGENT_MEMORY_EXTRACTION_PROMPT
        #      in mem0/mem0/configs/prompts.py to extract memories from
        #      messages containing role of "assistant", if agent_id is
        #      provided, otherwise it will use the
        #      USER_MEMORY_EXTRACTION_PROMPT in mem0/mem0/configs/prompts.py
        #      to extract memories.
        #
        # 3. Last resort: Record as "assistant" with infer=False
        #    - Used when both previous attempts yield no results
        #    - Bypasses mem0's inference mechanism, which means no
        #      inference is performed, mem0 will only record the content
        #      as is.
        #
        # This graduated approach ensures that even if mem0's inference fails
        # to extract meaningful memories, the raw content is still preserved.

        try:
            if thinking:
                content = [thinking] + content

            # Strategy 1: Record as user message first
            results = await self._mem0_record(
                [
                    {
                        "role": "user",
                        "content": "\n".join(content),
                        "name": "user",
                    },
                ],
                **kwargs,
            )

            # Strategy 2: Fallback to assistant message. In this case, if
            # agent_id is provided, mem0 will use the
            # AGENT_MEMORY_EXTRACTION_PROMPT in mem0/mem0/configs/prompts.py
            # to extract memories from messages containing role of
            # "assistant". If agent_id is not provided, mem0 will still use
            # the USER_MEMORY_EXTRACTION_PROMPT in
            # mem0/mem0/configs/prompts.py to extract memories.
            if (
                results
                and isinstance(results, dict)
                and "results" in results
                and len(results["results"]) == 0
            ):
                results = await self._mem0_record(
                    [
                        {
                            "role": "assistant",
                            "content": "\n".join(content),
                            "name": "assistant",
                        },
                    ],
                    **kwargs,
                )

            # Strategy 3: Last resort - direct recording without inference.
            # In this case, mem0 will not use any prompts to extract
            # memories, it will only record the content as is.
            if (
                results
                and isinstance(results, dict)
                and "results" in results
                and len(results["results"]) == 0
            ):
                results = await self._mem0_record(
                    [
                        {
                            "role": "assistant",
                            "content": "\n".join(content),
                            "name": "assistant",
                        },
                    ],
                    infer=False,
                    **kwargs,
                )

            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Successfully recorded content to memory "
                        f"{results}",
                    ),
                ],
            )

        except Exception as e:
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Error recording memory: {str(e)}",
                    ),
                ],
            )

    async def retrieve_from_memory(
        self,
        keywords: list[str],
        limit: int = 5,
        **kwargs: Any,
    ) -> ToolResponse:
        """Retrieve the memory based on the given keywords.

        Args:
            keywords (`list[str]`):
                The keywords to search for in the memory, which should be
                specific and concise, e.g. the person's name, the date, the
                location, etc.
            limit (`int`, optional):
                The maximum number of memories to retrieve per search.

        Returns:
            `ToolResponse`:
                A ToolResponse containing the retrieved memories as JSON text.
        """

        try:
            results = []
            for keyword in keywords:
                result = await self.long_term_working_memory.search(
                    query=keyword,
                    agent_id=self.agent_id,
                    user_id=self.user_id,
                    run_id=self.run_id,
                    limit=limit,
                )
                if result:
                    results.extend(
                        [item["memory"] for item in result["results"]],
                    )

            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text="\n".join(results),
                    ),
                ],
            )

        except Exception as e:
            return ToolResponse(
                content=[
                    TextBlock(
                        type="text",
                        text=f"Error retrieving memory: {str(e)}",
                    ),
                ],
            )

    async def record(
        self,
        msgs: list[Msg | None],
        memory_type: str | None = None,
        infer: bool = True,
        **kwargs: Any,
    ) -> None:
        """Record the content to the long-term memory.

        Args:
            msgs (`list[Msg | None]`):
                The messages to record to memory.
            memory_type (`str | None`, optional):
                The type of memory to use. Default is None, to create a
                semantic memory. "procedural_memory" is explicitly used for
                procedural memories.
            infer (`bool`, optional):
                Whether to infer memory from the content. Default is True.
            **kwargs (`Any`):
                Additional keyword arguments for the mem0 recording.
        """
        if isinstance(msgs, Msg):
            msgs = [msgs]

        # Filter out None
        msg_list = [_ for _ in msgs if _]
        if not all(isinstance(_, Msg) for _ in msg_list):
            raise TypeError(
                "The input messages must be a list of Msg objects.",
            )

        messages = [
            {
                "role": "assistant",
                "content": "\n".join([str(_.content) for _ in msg_list]),
                "name": "assistant",
            },
        ]

        await self._mem0_record(
            messages,
            memory_type=memory_type,
            infer=infer,
            **kwargs,
        )

    async def _mem0_record(
        self,
        messages: str | list[dict],
        memory_type: str | None = None,
        infer: bool = True,
        **kwargs: Any,
    ) -> dict:
        """Record the content to the long-term memory.

        Args:
            messages (`str`):
                The content to remember, which is a string or a list of
                dictionaries representing messages.
            memory_type (`str | None`, optional):
                The type of memory to use. Default is None, to create a
                semantic memory. "procedural_memory" is explicitly used for
                procedural memories.
            infer (`bool`, optional):
                Whether to infer memory from the content. Default is True.
            **kwargs (`Any`):
                Additional keyword arguments.

        Returns:
            `dict`:
                The result from the memory recording operation.
        """
        results = await self.long_term_working_memory.add(
            messages=messages,
            agent_id=self.agent_id,
            user_id=self.user_id,
            run_id=self.run_id,
            memory_type=(
                memory_type
                if memory_type is not None
                else self.default_memory_type
            ),
            infer=infer,
            **kwargs,
        )
        return results

    async def retrieve(
        self,
        msg: Msg | list[Msg] | None,
        limit: int = 5,
        **kwargs: Any,
    ) -> str:
        """Retrieve the content from the long-term memory.

        Args:
            msg (`Msg | list[Msg] | None`):
                The message to search for in the memory, which should be
                specific and concise, e.g. the person's name, the date, the
                location, etc.
            limit (`int`, optional):
                The maximum number of memories to retrieve per search.
            **kwargs (`Any`):
                Additional keyword arguments.

        Returns:
            `str`:
                The retrieved memory
        """
        if isinstance(msg, Msg):
            msg = [msg]

        if not isinstance(msg, list) or not all(
            isinstance(_, Msg) for _ in msg
        ):
            raise TypeError(
                "The input message must be a Msg or a list of Msg objects.",
            )

        msg_strs = [
            json.dumps(_.to_dict()["content"], ensure_ascii=False) for _ in msg
        ]

        results = []
        for item in msg_strs:
            result = await self.long_term_working_memory.search(
                query=item,
                agent_id=self.agent_id,
                user_id=self.user_id,
                run_id=self.run_id,
                limit=limit,
            )
            if result:
                results.extend([item["memory"] for item in result["results"]])

        return "\n".join(results)



# -*- coding: utf-8 -*-
"""The document data structure used in RAG as the data chunk and
retrieval result."""
from dataclasses import dataclass, field

import shortuuid
from dashscope.api_entities.dashscope_response import DictMixin

from ..message import (
    TextBlock,
    ImageBlock,
    VideoBlock,
)
from ..types import Embedding


@dataclass
class DocMetadata(DictMixin):
    """The metadata of the document."""

    content: TextBlock | ImageBlock | VideoBlock
    """The data content, e.g., text, image, video."""

    doc_id: str
    """The document ID."""

    chunk_id: int
    """The chunk ID."""

    total_chunks: int
    """The total number of chunks."""


@dataclass
class Document:
    """The data chunk."""

    metadata: DocMetadata
    """The metadata of the data chunk."""

    id: str = field(default_factory=shortuuid.uuid)
    """The unique ID of the data chunk."""

    # The fields that will be filled when the document is added to or
    # retrieved from the knowledge base.

    embedding: Embedding | None = field(default_factory=lambda: None)
    """The embedding of the data chunk."""

    score: float | None = None
    """The relevance score of the data chunk."""



# -*- coding: utf-8 -*-
"""The agentscope serialization module"""
import os

import requests

from . import exception
from . import module
from . import message
from . import model
from . import tool
from . import formatter
from . import memory
from . import agent
from . import session
from . import embedding
from . import token
from . import evaluate
from . import pipeline
from . import tracing
from . import rag

from ._logging import (
    logger,
    setup_logger,
)
from .hooks import _equip_as_studio_hooks
from ._version import __version__


def init(
    project: str | None = None,
    name: str | None = None,
    logging_path: str | None = None,
    logging_level: str = "INFO",
    studio_url: str | None = None,
    tracing_url: str | None = None,
) -> None:
    """Initialize the agentscope library.

    Args:
        project (`str | None`, optional):
            The project name.
        name (`str | None`, optional):
            The name of the run.
        logging_path (`str | None`, optional):
            The path to saving the log file. If not provided, logs will not be
            saved.
        logging_level (`str | None`, optional):
            The logging level. Defaults to "INFO".
        studio_url (`str | None`, optional):
            The URL of the AgentScope Studio to connect to.
        tracing_url (`str | None`, optional):
            The URL of the tracing endpoint, which can connect to third-party
            OpenTelemetry tracing platforms like Arize-Phoenix and Langfuse.
            If not provided and `studio_url` is provided, it will send traces
            to the AgentScope Studio's tracing endpoint.
    """

    from . import _config

    if project:
        _config.project = project

    if name:
        _config.name = name

    setup_logger(logging_level, logging_path)

    if studio_url:
        # Register the run
        data = {
            "id": _config.run_id,
            "project": _config.project,
            "name": _config.name,
            "timestamp": _config.created_at,
            "pid": os.getpid(),
            "status": "running",
            # Deprecated fields
            "run_dir": "",
        }
        response = requests.post(
            url=f"{studio_url}/trpc/registerRun",
            json=data,
        )
        response.raise_for_status()

        from .agent import UserAgent, StudioUserInput

        UserAgent.override_class_input_method(
            StudioUserInput(
                studio_url=studio_url,
                run_id=_config.run_id,
                max_retries=3,
            ),
        )

        _equip_as_studio_hooks(studio_url)

    if tracing_url:
        endpoint = tracing_url
    else:
        endpoint = studio_url.strip("/") + "/v1/traces" if studio_url else None

    if endpoint:
        from .tracing import setup_tracing

        setup_tracing(endpoint=endpoint)


__all__ = [
    # modules
    "exception",
    "module",
    "message",
    "model",
    "tool",
    "formatter",
    "memory",
    "agent",
    "session",
    "logger",
    "embedding",
    "token",
    "evaluate",
    "pipeline",
    "tracing",
    "rag",
    # functions
    "init",
    "setup_logger",
    "__version__",
]
